/*
 * GStreamer
 * Copyright (C) 2020 Igalia S.L.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#ifndef _GST_EGL_IMAGE_PRIVATE_H_
#define _GST_EGL_IMAGE_PRIVATE_H_

#include <gst/gl/gstgl_fwd.h>
#include <gst/gl/gstglformat.h>

G_BEGIN_DECLS

G_GNUC_INTERNAL
gboolean                gst_egl_image_check_dmabuf_direct       (GstGLContext * context,
                                                                 const GstVideoInfo * in_info,
                                                                 GstGLTextureTarget target);

G_GNUC_INTERNAL
gboolean                gst_egl_image_check_dmabuf_direct_with_dma_drm
                                                                (GstGLContext * context,
                                                                 const GstVideoInfoDmaDrm * in_info_dma,
                                                                 GstGLTextureTarget target);

G_END_DECLS

#endif /* _GST_EGL_IMAGE_PRIVATE_H_ */
