
GstCheckABIStruct list[] = {
  {"GstAppSinkCallbacks", sizeof (GstAppSinkCallbacks), 28},
  {"GstAppSinkClass", sizeof (GstAppSinkClass), 436},
  {"GstAppSink", sizeof (GstAppSink), 460},
  {"GstAppSrcCallbacks", sizeof (GstAppSrcCallbacks), 28},
  {"GstAppSrcClass", sizeof (GstAppSrcClass), 440},
  {"GstAppSrc", sizeof (GstAppSrc), 444},
  {"GstAudioCdSrcClass", sizeof (GstAudioCdSrcClass), 524},
  {"GstAudioCdSrc", sizeof (GstAudioCdSrc), 464},
  {"GstAudioCdSrcTrack", sizeof (GstAudioCdSrcTrack), 36},
  {"GstAudioClockClass", sizeof (GstAudioClockClass), 164},
  {"GstAudioClock", sizeof (GstAudioClock), 140},
  {"GstAudioFilterClass", sizeof (GstAudioFilterClass), 436},
  {"GstAudioFilter", sizeof (GstAudioFilter), 676},
  {"GstAudioSinkClass", sizeof (GstAudioSinkClass), 468},
  {"GstAudioSink", sizeof (GstAudioSink), 516},
  {"GstAudioSrcClass", sizeof (GstAudioSrcClass), 496},
  {"GstAudioSrc", sizeof (GstAudioSrc), 512},
  {"GstAudioBaseSinkClass", sizeof (GstAudioBaseSinkClass), 424},
  {"GstAudioBaseSink", sizeof (GstAudioBaseSink), 496},
  {"GstAudioBaseSrcClass", sizeof (GstAudioBaseSrcClass), 452},
  {"GstAudioBaseSrc", sizeof (GstAudioBaseSrc), 492},
  {"GstRTPBaseAudioPayloadClass", sizeof (GstRTPBaseAudioPayloadClass), 304},
  {"GstRTPBaseAudioPayload", sizeof (GstRTPBaseAudioPayload), 420},
  {"GstRTPBaseDepayloadClass", sizeof (GstRTPBaseDepayloadClass), 280},
  {"GstRTPBaseDepayload", sizeof (GstRTPBaseDepayload), 304},
  {"GstRTPBasePayloadClass", sizeof (GstRTPBasePayloadClass), 288},
  {"GstRTPBasePayload", sizeof (GstRTPBasePayload), 380},
  {"GstColorBalanceChannelClass", sizeof (GstColorBalanceChannelClass), 88},
  {"GstColorBalanceChannel", sizeof (GstColorBalanceChannel), 40},
  {"GstColorBalanceInterface", sizeof (GstColorBalanceInterface), 44},
  {"GstDiscovererClass", sizeof (GstDiscovererClass), 100},
  {"GstDiscoverer", sizeof (GstDiscoverer), 32},
  {"GstFFTF32Complex", sizeof (GstFFTF32Complex), 8},
  {"GstFFTF64Complex", sizeof (GstFFTF64Complex), 16},
  {"GstFFTS16Complex", sizeof (GstFFTS16Complex), 4},
  {"GstFFTS32Complex", sizeof (GstFFTS32Complex), 8},
  {"GstNavigationInterface", sizeof (GstNavigationInterface), 16},
  {"gst_riff_acid", sizeof (gst_riff_acid), 24},
  {"gst_riff_dmlh", sizeof (gst_riff_dmlh), 4},
  {"gst_riff_index_entry", sizeof (gst_riff_index_entry), 16},
  {"gst_riff_strf_auds", sizeof (gst_riff_strf_auds), 16},
  {"gst_riff_strf_iavs", sizeof (gst_riff_strf_iavs), 32},
  {"gst_riff_strf_vids", sizeof (gst_riff_strf_vids), 40},
  {"gst_riff_strh", sizeof (gst_riff_strh), 48},
  {"GstAudioRingBufferClass", sizeof (GstAudioRingBufferClass), 156},
  {"GstAudioRingBuffer", sizeof (GstAudioRingBuffer), 756},
  {"GstAudioRingBufferSpec", sizeof (GstAudioRingBufferSpec), 348},
  {"GstRTCPPacket", sizeof (GstRTCPPacket), 36},
  {"GstRTPPayloadInfo", sizeof (GstRTPPayloadInfo), 40},
  {"GstRTSPExtensionInterface", sizeof (GstRTSPExtensionInterface), 64},
  {"GstRTSPMessage", sizeof (GstRTSPMessage), 44},
  {"GstRTSPRange", sizeof (GstRTSPRange), 8},
  {"GstRTSPTime", sizeof (GstRTSPTime), 12},
  {"GstRTSPTimeRange", sizeof (GstRTSPTimeRange), 68},
  {"GstRTSPTransport", sizeof (GstRTSPTransport), 92},
  {"GstRTSPUrl", sizeof (GstRTSPUrl), 32},
  {"GstRTSPWatchFuncs", sizeof (GstRTSPWatchFuncs), 48},
  {"GstSDPAttribute", sizeof (GstSDPAttribute), 8},
  {"GstSDPBandwidth", sizeof (GstSDPBandwidth), 8},
  {"GstSDPConnection", sizeof (GstSDPConnection), 20},
  {"GstSDPKey", sizeof (GstSDPKey), 8},
  {"GstSDPMedia", sizeof (GstSDPMedia), 44},
  {"GstSDPMessage", sizeof (GstSDPMessage), 96},
  {"GstSDPOrigin", sizeof (GstSDPOrigin), 24},
  {"GstSDPTime", sizeof (GstSDPTime), 12},
  {"GstSDPZone", sizeof (GstSDPZone), 8},
  {"GstStreamVolumeInterface", sizeof (GstStreamVolumeInterface), 8},
  {"GstTagDemuxClass", sizeof (GstTagDemuxClass), 284},
  {"GstTagDemux", sizeof (GstTagDemux), 192},
  {"GstVideoFilterClass", sizeof (GstVideoFilterClass), 444},
  {"GstVideoFilter", sizeof (GstVideoFilter), 608},
  {"GstVideoOrientationInterface", sizeof (GstVideoOrientationInterface), 40},
  {"GstVideoRectangle", sizeof (GstVideoRectangle), 16},
  {"GstVideoSinkClass", sizeof (GstVideoSinkClass), 420},
  {"GstVideoSink", sizeof (GstVideoSink), 468},
  {"GstVideoOverlayInterface", sizeof (GstVideoOverlayInterface), 24},
#if defined(TEST_GST_GL_ABI_CHECK)
  {"GstGLAsyncDebug", sizeof (GstGLAsyncDebug), 120},
  {"GstGLBaseFilter", sizeof (GstGLBaseFilter), 648},
  {"GstGLBaseFilterClass", sizeof (GstGLBaseFilterClass), 880},
  {"GstGLBaseMemory", sizeof (GstGLBaseMemory), 224},
  {"GstGLAllocationParams", sizeof (GstGLAllocationParams), 120},
  {"GstGLBaseMemoryAllocator", sizeof (GstGLBaseMemoryAllocator), 216},
  {"GstGLBaseMemoryAllocatorClass", sizeof (GstGLBaseMemoryAllocatorClass), 312},
  {"GstGLBaseSrc", sizeof (GstGLBaseSrc), 904},
  {"GstGLBaseSrcClass", sizeof (GstGLBaseSrcClass), 920},
  {"GstGLBuffer", sizeof (GstGLBuffer), 240},
  {"GstGLBufferAllocationParams", sizeof (GstGLBufferAllocationParams), 160},
  {"GstGLBufferAllocator", sizeof (GstGLBufferAllocator), 248},
  {"GstGLBufferAllocatorClass", sizeof (GstGLBufferAllocatorClass), 344},
  {"GstGLBufferPool", sizeof (GstGLBufferPool), 184},
  {"GstGLBufferPoolClass", sizeof (GstGLBufferPoolClass), 320},
  {"GstGLColorConvert", sizeof (GstGLColorConvert), 480},
  {"GstGLColorConvertClass", sizeof (GstGLColorConvertClass), 216},
  {"GstGLContext", sizeof (GstGLContext), 152},
  {"GstGLContextClass", sizeof (GstGLContextClass), 312},
  {"GstGLDisplay", sizeof (GstGLDisplay), 136},
  {"GstGLDisplayClass", sizeof (GstGLDisplayClass), 232},
  {"GstGLFilter", sizeof (GstGLFilter), 1064},
  {"GstGLFilterClass", sizeof (GstGLFilterClass), 952},
  {"GstGLFuncs", sizeof (GstGLFuncs), 2968},
  {"GstGLFramebuffer", sizeof (GstGLFramebuffer), 152},
  {"GstGLFramebufferClass", sizeof (GstGLFramebufferClass), 216},
  {"GstGLMemory", sizeof (GstGLMemory), 480},
  {"GstGLVideoAllocationParams", sizeof (GstGLVideoAllocationParams), 184},
  {"GstGLMemoryAllocator", sizeof (GstGLMemoryAllocator), 248},
  {"GstGLMemoryAllocatorClass", sizeof (GstGLMemoryAllocatorClass), 368},
  {"GstGLMemoryPBO", sizeof (GstGLMemoryPBO), 520},
  {"GstGLMemoryPBOAllocator", sizeof (GstGLMemoryPBOAllocator), 280},
  {"GstGLMemoryPBOAllocatorClass", sizeof (GstGLMemoryPBOAllocatorClass), 400},
  {"GstGLRenderbuffer", sizeof (GstGLRenderbuffer), 280},
  {"GstGLRenderbufferAllocationParams", sizeof (GstGLRenderbufferAllocationParams), 168},
  {"GstGLRenderbufferAllocator", sizeof (GstGLRenderbufferAllocator), 248},
  {"GstGLRenderbufferAllocatorClass", sizeof (GstGLRenderbufferAllocatorClass), 344},
  {"GstGLSLStage", sizeof (GstGLSLStage), 136},
  {"GstGLSLStageClass", sizeof (GstGLSLStageClass), 216},
  {"GstGLShader", sizeof (GstGLShader), 136},
  {"GstGLShaderClass", sizeof (GstGLShaderClass), 184},
  {"GstGLUpload", sizeof (GstGLUpload), 136},
  {"GstGLUploadClass", sizeof (GstGLUploadClass), 216},
  {"GstGLOverlayCompositor", sizeof (GstGLOverlayCompositor), 160},
  {"GstGLOverlayCompositorClass", sizeof (GstGLOverlayCompositorClass), 216},
  {"GstGLQuery", sizeof (GstGLQuery), 176},
  {"GstGLViewConvert", sizeof (GstGLViewConvert), 504},
  {"GstGLViewConvertClass", sizeof (GstGLViewConvertClass), 216},
  {"GstGLWindow", sizeof (GstGLWindow), 248},
  {"GstGLWindowClass", sizeof (GstGLWindowClass), 336},
#if GST_GL_HAVE_PLATFORM_EGL
  {"GstGLDisplayEGL", sizeof (GstGLDisplayEGL), 184},
  {"GstGLDisplayEGLClass", sizeof (GstGLDisplayEGLClass), 264},
  {"GstGLDisplayEGLDevice", sizeof (GstGLDisplayEGL), 184},
  {"GstGLDisplayEGLDeviceClass", sizeof (GstGLDisplayEGLClass), 264},
  {"GstGLMemoryEGL", sizeof (GstGLMemoryEGL), 520},
  {"GstGLMemoryEGLAllocator", sizeof (GstGLMemoryEGLAllocator), 280},
  {"GstGLMemoryEGLAllocatorClass", sizeof (GstGLMemoryEGLAllocatorClass), 400},
  {"GstEGLImage", sizeof (GstGLMemoryEGLAllocatorClass), 400},
#endif
#if GST_GL_HAVE_WINDOW_WAYLAND
  {"GstGLDisplayWayland", sizeof (GstGLDisplayWayland), 216},
  {"GstGLDisplayWaylandClass", sizeof (GstGLDisplayWaylandClass), 264},
#endif
#if GST_GL_HAVE_WINDOW_X11
  {"GstGLDisplayX11", sizeof (GstGLDisplayX11), 200},
  {"GstGLDisplayX11Class", sizeof (GstGLDisplayX11Class), 264},
#endif
#endif
  {NULL, 0, 0}
};
